# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import
from heptapod.testhelpers import LocalRepoWrapper

from .utils import common_config


def test_initial_import_settings(tmpdir):
    config = common_config()
    config['heptapod'] = {'initial-import': 'yes'}
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'), config=config)

    ui = wrapper.repo.ui
    assert not ui.configbool(b'experimental', b'single-head-per-branch')
    assert not ui.configbool(b'experimental', b'topic.publish-bare-branch')



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Heptapod specific branch facilities.

Default GitLab branch:
  This a GitLab branch, e.g., ``branch/default`` rather than a Mercurial
  branch, e.g., ``default``.

  It is stored inside the repository directory, but not in the main store.
  In principle, different shares could have different default GitLab branches,
  if that were to be useful.

  For comparison, GitLab uses the value of `HEAD` on Git repositories for
  this.
"""
from mercurial.i18n import _
from mercurial import error

DEFAULT_GITLAB_BRANCH_FILE_NAME = b'default_gitlab_branch'


def get_default_gitlab_branch(repo):
    """Return the default GitLab branch name, or ``None`` if not set."""
    branch = repo.vfs.tryread(DEFAULT_GITLAB_BRANCH_FILE_NAME)
    # (hg 5.4) tryread returns empty strings for missing files
    if not branch:
        return None
    return branch


def set_default_gitlab_branch(repo, target):
    if not target:
        raise error.Abort(_("The default GitLab branch cannot be an "
                            "empty string."))
    with repo.wlock():
        repo.vfs.write(DEFAULT_GITLAB_BRANCH_FILE_NAME, target)



# Copyright 2019 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import
from mercurial import (
    error,
    extensions,
    phases,
    ui as uimod,
)
from os.path import dirname
import pytest

from ..testhelpers import (
    LocalRepoWrapper,
    NULL_ID,
    NULL_REVISION,
)
import hgext3rd.heptapod

HGEXT_HEPTA_SOURCE = dirname(hgext3rd.heptapod.__file__)


parametrize = pytest.mark.parametrize


def test_init_write_commit(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir)
    node = wrapper.write_commit('foo', content='Foo', message='Foo committed')
    ctx = wrapper.repo[node]
    assert ctx.description() == b'Foo committed'
    parents = ctx.parents()
    assert len(parents) == 1
    assert parents[0].rev() == NULL_REVISION
    assert parents[0].node() == NULL_ID

    del wrapper, ctx

    reloaded = LocalRepoWrapper.load(tmpdir)
    rl_ctx = reloaded.repo[node]
    assert rl_ctx.description() == b'Foo committed'


def assert_is_hepta_ext(hepta_ext):
    assert hepta_ext is not None
    # it's imported with a different name, hence can't be directly compared
    # let's also avoid flakiness due to __file__ behaviour depending on
    # installation context
    assert hepta_ext.__doc__ == hgext3rd.heptapod.__doc__


def test_load_hgrc_extension(tmpdir):
    LocalRepoWrapper.init(tmpdir)
    tmpdir.join('.hg', 'hgrc').write('\n'.join((
        "[extensions]",
        "heptapod=" + HGEXT_HEPTA_SOURCE,
    )))
    wrapper = LocalRepoWrapper.load(tmpdir, config=dict(foo=dict(bar='17')))
    exts = dict(extensions.extensions(wrapper.repo.ui))
    assert_is_hepta_ext(exts.get(b'heptapod'))

    assert wrapper.repo.ui.configint(b'foo', b'bar') == 17


def test_init_baseui_config_extension(tmpdir):
    ui = uimod.ui.load()
    ui.setconfig(b'foo', b'bar', b'yes', source=b'tests')
    ui.setconfig(b'extensions', b'heptapod', HGEXT_HEPTA_SOURCE.encode(),
                 source=b'tests')
    wrapper = LocalRepoWrapper.init(tmpdir, base_ui=ui)

    assert wrapper.repo.ui.configbool(b'foo', b'bar')
    exts = dict(extensions.extensions(wrapper.repo.ui))
    assert_is_hepta_ext(exts.get(b'heptapod'))


def test_init_config_extension(tmpdir):
    ui = uimod.ui.load()
    ui.setconfig(b'foo', b'bar', b'yes', source=b'tests')
    ui.setconfig(b'extensions', b'heptapod', HGEXT_HEPTA_SOURCE.encode(),
                 source=b'tests')
    wrapper = LocalRepoWrapper.init(
        tmpdir,
        config=dict(foo=dict(bar='yes'),
                    extensions=dict(heptapod=HGEXT_HEPTA_SOURCE),
                    ))

    assert wrapper.repo.ui.configbool(b'foo', b'bar')
    exts = dict(extensions.extensions(wrapper.repo.ui))
    assert_is_hepta_ext(exts.get(b'heptapod'))


def test_update(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir)
    wrapper.write_commit('foo', content='Foo 0')
    node1 = wrapper.write_commit('foo', content='Foo 1', return_ctx=False)
    foo = tmpdir.join('foo')
    assert foo.read() == 'Foo 1'

    wrapper.update('0')
    assert foo.read() == 'Foo 0'

    wrapper.update_bin(NULL_ID)
    assert not foo.isfile()

    wrapper.update_bin(node1)
    assert foo.read() == 'Foo 1'


@parametrize('meth', ['class', 'instance'])
def test_share(tmpdir, meth):
    main_path = tmpdir.join('main')
    dest_path = tmpdir.join('dest')
    main_wrapper = LocalRepoWrapper.init(main_path)
    node0 = main_wrapper.write_commit('foo', message='Done in main')

    if meth == 'class':
        dest_wrapper = LocalRepoWrapper.share_from_path(main_path, dest_path)
    else:
        dest_wrapper = main_wrapper.share(dest_path)

    dest_ctx = dest_wrapper.repo[node0]
    assert dest_ctx.description() == b'Done in main'

    node1 = dest_wrapper.write_commit(
        'foo', message='Done in dest', parent=node0)

    # we need to reload the main repo to see the new changeset
    reloaded = LocalRepoWrapper.load(main_path)
    main_ctx = reloaded.repo[node1]
    assert main_ctx.description() == b'Done in dest'


def test_write_commit_named_branch(tmpdir):
    """Demonstrate the use of write_commit with parent."""
    wrapper = LocalRepoWrapper.init(tmpdir)
    ctx0 = wrapper.write_commit('foo', content='Foo 0')
    wrapper.write_commit('foo', content='Foo 1')
    ctxbr = wrapper.write_commit('foo', content='Foo branch',
                                 parent=ctx0, branch='other')

    assert ctxbr.branch() == b'other'
    assert ctxbr.parents() == [ctx0]


def test_write_commit_parent_nodeid(tmpdir):
    """Demonstrate the use of write_commit with parent given as node id"""
    wrapper = LocalRepoWrapper.init(tmpdir)
    node0 = wrapper.write_commit('foo', content='Foo 0', return_ctx=False)
    assert isinstance(node0, bytes)  # avoid tautologies

    wrapper.write_commit('foo', content='Foo 1')
    ctxbr = wrapper.write_commit('foo', content='Foo branch',
                                 parent=node0, branch='other')

    assert ctxbr.branch() == b'other'
    assert [c.node() for c in ctxbr.parents()] == [node0]


def test_write_commit_time(tmpdir):
    """Demonstrate the utc_timestamp optional parameter."""
    wrapper = LocalRepoWrapper.init(tmpdir)
    ctx = wrapper.write_commit('foo', content='Foo', utc_timestamp=123456)
    assert ctx.date() == (123456.0, 0)

    # floats are accepted, but get truncated
    ctx = wrapper.write_commit('foo', content='Foo2', utc_timestamp=12.34)
    assert ctx.date() == (12.0, 0)


def test_write_commit_topic(tmpdir):
    """Demonstrate the use of write_commit with parent."""
    # it is essential to activate the rebase extension, even though
    # we don't use it in this test, because the
    # first loading of topic patches it only if it is present.
    # without this, all subsequent tests expecting rebase to preserve
    # topics would be broken
    wrapper = LocalRepoWrapper.init(tmpdir,
                                    config=dict(extensions=dict(rebase='',
                                                                topic='')))

    ctx0 = wrapper.write_commit('foo', content='Foo 0')
    wrapper.write_commit('foo', content='Foo 1')
    ctxbr = wrapper.write_commit('foo', content='Foo branch',
                                 parent=ctx0, topic='sometop')

    assert ctxbr.topic() == b'sometop'
    assert ctxbr.parents() == [ctx0]


def test_write_commit_wild_branch(tmpdir):
    """Demonstrate the use of write_commit with parent."""
    wrapper = LocalRepoWrapper.init(tmpdir)
    ctx0 = wrapper.write_commit('foo', content='Foo 0')
    wrapper.write_commit('foo', content='Foo 1')
    ctxbr = wrapper.write_commit('foo', content='Foo branch',
                                 parent=ctx0)

    assert ctxbr.branch() == b'default'
    assert ctxbr.parents() == [ctx0]


def test_write_commit_random(tmpdir):
    """Demonstrate how random content is generated."""

    wrapper = LocalRepoWrapper.init(tmpdir)
    node0 = wrapper.write_commit('foo')
    ctx1 = wrapper.write_commit('foo', parent=node0, return_ctx=True)
    ctx2 = wrapper.write_commit('foo', parent=node0, return_ctx=True)

    assert ctx1.p1() == ctx2.p1()
    assert ctx1 != ctx2


def test_phase(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir)
    node = wrapper.write_commit('foo', content='Foo 0')
    ctx = wrapper.repo[node]
    assert ctx.phase() == phases.draft

    wrapper.set_phase('public', ['.'], force=False)
    assert ctx.phase() == phases.public

    wrapper.set_phase('draft', ['.'], force=True)
    assert ctx.phase() == phases.draft


def test_prune_update_hidden(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir,
                                    config=dict(extensions=dict(evolve='')))
    wrapper.write_commit('foo', content='Foo 0')
    ctx = wrapper.write_commit('foo', content='Foo 1', return_ctx=True)
    wrapper.prune('.')
    assert ctx.obsolete()

    wrapper.update(0)
    assert tmpdir.join('foo').read() == 'Foo 0'

    with pytest.raises(error.FilteredRepoLookupError):
        wrapper.update(ctx.hex())

    wrapper.update_bin(ctx.node(), hidden=True)
    assert tmpdir.join('foo').read() == 'Foo 1'



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Inner tests for the Git subsystem.

These wouldn't really qualify as unit tests, since they exert at least
Mercurial, but they are more unitary and disposable than those of
test_integration, testing the implementation details of the Git subsystem.
"""
from __future__ import absolute_import

from mercurial import (
    error,
)
import pytest

from heptapod.testhelpers import (
    LocalRepoWrapper,
)
from heptapod.gitlab import prune_reasons
from ...git import (
    git_branch_ref,
    HeptapodGitHandler,
)
from mercurial import pycompat
from ..utils import common_config


def bkey(**d):
    """Replacement for `dict(k=v, ...)` that produces bytes keys"""
    return pycompat.byteskwargs(d)


def test_heptapod_gate_bookmarks(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'),
                                    config=common_config())
    ctx = wrapper.write_commit('foo', content='foo0', message="default0",
                               return_ctx=True)

    repo = wrapper.repo
    handler = HeptapodGitHandler(repo, repo.ui)
    gate = handler.heptapod_gate_bookmarks

    assert len(gate(repo, True, {})) == 0

    def book_changes(**changes):
        """Readability helper because of bytes-everywhere"""
        # keys have to be byteified
        return {pycompat.sysbytes(book): change
                for book, change in changes.items()}

    with pytest.raises(error.Abort) as exc_info:
        gate(repo, False, book_changes(book1=(None, '\x00\x01')))
    assert b'forbidden' in exc_info.value.args[0]
    assert b'book1' in exc_info.value.args[0]

    deleted = gate(repo, False, book_changes(book1=('\x00\x01', None)))
    assert list(deleted) == [b'book1']

    deleted = gate(repo, True, book_changes(book1=(None, ctx.node())))
    assert not deleted

    wrapper.command('topics', b'zetop')
    node = wrapper.write_commit('zz', message='topical')
    with pytest.raises(error.Abort) as exc_info:
        gate(repo, True, book_changes(book1=(None, node)))
    assert b'forbidden' in exc_info.value.args[0]
    assert b'topic' in exc_info.value.args[0]


def test_git_branch_for_branchmap_branch(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'))
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    gb = handler.git_branch_for_branchmap_branch

    assert gb(b'default') == b'branch/default'
    assert gb(b'default:zztop') == b'topic/default/zztop'

    with pytest.raises(error.Abort) as exc_info:
        assert gb(b'default:zz/top')
    assert b"Invalid character '/'" in exc_info.value.args[0]

    wrapper.repo.ui.setconfig(b'experimental',
                              b'hg-git.accept-slash-in-topic-name', True)

    assert gb(b'default:zz/top') == b'topic/default/zz-top'


def test_multiple_heads_choose_corner_case(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'))
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    assert handler.multiple_heads_choose((), 'branch name') is None


def test_notify_gitlab_empty_changes(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'))
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    assert handler.heptapod_notify_gitlab('some-hook', {}, {}) is None


def test_analyse_vanished_topic_lookup_error(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'))
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    unknown_sha = b'1234beef' * 5
    prune = handler.analyse_vanished_topic(
        b'default', b'zetop', unknown_sha,
        log_info=bkey(ref=b'refs/heads/topic/default/zetop',
                      before_git_sha=b'01234cafe'))
    assert prune == prune_reasons.HeadPruneReason()


def test_analyse_vanished_bogus_topic(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'))
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    assert handler.analyze_vanished_refs(
        {b'refs/heads/topic/missing-branch-name': b'ca1234fe',
         b'refs/heads/branch/default': 'will be completely ignored anyway'},
        {b'refs/heads/branch/default': 'will be completely ignored anyway'},
    ) == {}


def test_analyse_vanished_topic_draft_succ_unknown_reason(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'),
                                    config=common_config())
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    wrapper.write_commit('foo', message='foo0')
    ctx = wrapper.write_commit('foo', message='in topic', topic='zzetop',
                               return_ctx=True)

    # Normally, analyse_vanished_topic works under the assumption that
    # the topic is not visible anymore. Yet feeding a visible topic
    # is the only currently known way of triggering the safety return
    # of before_sha for unknown situations

    prune = handler.analyse_vanished_topic(
        b'default', b'zzetop',
        ctx.hex(),
        log_info=bkey(ref='refs/heads/topic/default/zetop',
                      before_git_sha='01234cafe'))

    # TODO me way use a more precise reason later on
    assert prune == prune_reasons.HeadPruneReason()


def test_analyse_vanished_topic_latest_succ_not_found(tmpdir, monkeypatch):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'),
                                    config=common_config())
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    wrapper.write_commit('foo', message='foo0')
    ctx = wrapper.write_commit('foo', message='in topic', topic='zzetop',
                               return_ctx=True)
    wrapper.command('amend', message=b'amended')
    wrapper.set_phase('public', ['.'])

    monkeypatch.setattr(handler, 'published_topic_latest_hg_sha',
                        lambda *a, **kw: None)

    prune = handler.analyse_vanished_topic(
        b'default', b'zzetop',
        ctx.hex(),
        log_info=bkey(ref='refs/heads/topic/default/zetop',
                      before_git_sha='01234cafe'))
    # TODO me way use a more precise reason later on
    assert prune == prune_reasons.HeadPruneReason()


def prepare_topic_divergence(tmpdir, additional_config=None):
    config = common_config()
    config['experimental'] = {'evolution.allowdivergence': 'yes'}
    if additional_config is not None:
        config.update(additional_config)
    repo_path = tmpdir.join('repo')
    wrapper = LocalRepoWrapper.init(repo_path, config=config)

    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    wrapper.write_commit('foo', content='foo0', message="default0")
    ctx = wrapper.write_commit('foo', content='obs', message="to obsolete",
                               return_ctx=True)

    # create a content divergence
    repo_path.join('foo').write('amend1')
    wrapper.command('amend', message=b'amend1')

    wrapper.update(ctx.hex(), hidden=True)
    repo_path.join('foo').write('amend2')
    wrapper.command('amend', message=b'amend2')

    return wrapper, ctx, handler


def test_analyse_vanished_topic_divergence_initial_import(tmpdir):
    wrapper, ctx, handler = prepare_topic_divergence(
        tmpdir,
        additional_config=dict(heptapod={'initial-import': True}))

    handler.analyse_vanished_topic(
        'default', 'zetop',
        ctx.hex(),
        log_info=dict(ref='refs/heads/topic/default/zetop',
                      before_git_sha='01234cafe'))


def test_analyse_vanished_topic_divergence_not_initial_import(tmpdir):
    wrapper, ctx, handler = prepare_topic_divergence(tmpdir)

    with pytest.raises(error.Abort) as exc_info:
        handler.analyse_vanished_topic(
            'default', 'zetop',
            ctx.hex(),
            log_info={b'ref': b'refs/heads/topic/default/zetop',
                      b'before_git_sha': b'01234cafe'})

    # TODO UPSTREAM evolve gives a str message instead of the standard
    # bytes for error.Abort
    exc_msg = pycompat.sysbytes(exc_info.value.args[0])
    assert ctx.hex() in exc_msg
    assert b'divergent' in exc_msg


def test_topic_published_multiple_heads(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'),
                                    config=common_config())
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    wrapper.write_commit('foo', message='foo0')
    before_ctx = wrapper.write_commit('foo', message='in topic',
                                      topic='zzetop', return_ctx=True)

    # let's publish for the sake of consistency, but these inner tests
    # aren't based on the state of a Git repo anyway: previous state is
    # passed explicitely
    wrapper.set_phase('public', ['zzetop'])

    # let's add two public in topic descendants
    ctx1 = wrapper.write_commit('foo', message='this is wild 1',
                                parent=before_ctx.node(), topic='zzetop',
                                return_ctx=True)
    ctx2 = wrapper.write_commit('foo', message='this is wild 2',
                                parent=before_ctx.node(), topic='zzetop',
                                return_ctx=True)
    wrapper.set_phase('public', [ctx1.hex(), ctx2.hex()])

    before_git_sha = b'01234cafe'
    handler.map_set(before_git_sha, before_ctx.hex())
    prune = handler.analyse_vanished_topic(
        b'default', b'zzetop',
        before_ctx.hex(),
        log_info=dict(ref='refs/heads/topic/default/zzetop',
                      before_git_sha=before_git_sha))
    assert prune == prune_reasons.TopicPublished(before_git_sha)


def test_latest_unique_successor_no_descendant(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'),
                                    config=common_config())
    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    ctx = wrapper.write_commit('foo', message='foo0', return_ctx=True)
    assert handler.latest_topic_descendant(b'sometop', ctx) is None


def test_generate_prune_changes_existing_unknown(tmpdir):
    config = common_config()
    config['experimental'] = {'hg-git.prune-previously-closed-branches': 'no'}
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'), config=config)

    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)

    # let's have a git -> hg map that gives us an unknown hg sha.
    # it still has to be formally valid, though (hex, length 40)
    map_hg = {b'1234': b'cafe' * 10}

    def map_hg_get(git_sha):
        return map_hg.get(git_sha)

    handler.map_hg_get = map_hg_get

    gitlab_branch = b'branch/previously-closed'
    branch_ref = git_branch_ref(gitlab_branch)
    existing = {branch_ref: b'1234'}
    to_prune = {gitlab_branch: 'closed'}
    closed_reason = prune_reasons.BranchClosed()

    reasons, changes = handler.generate_prune_changes(to_prune, existing)
    assert reasons == {gitlab_branch: closed_reason}

    # and now with the previous Git changeset not even known
    existing = {branch_ref: b'unknown'}
    reasons, changes = handler.generate_prune_changes(to_prune, existing)
    assert reasons == {gitlab_branch: closed_reason}


def test_heptapod_compare_tags_inconsistent(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'),
                                    config=common_config())
    wrapper.write_commit('foo')
    # yes '..' is a valid tag name in Mercurial, no wonder Git doesn't like it
    wrapper.command('tag', b'0.12.3', rev=b'.')

    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)

    # the handler's hg_sha -> git_sha mapping being empty, it won't find the
    # Git SHA for our changeset, and that's the case we're testing
    changes = handler.heptapod_compare_tags({})
    assert changes == {}


def test_heptapod_compare_tags_invalid_git_name(tmpdir):
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'),
                                    config=common_config())
    ctx = wrapper.write_commit('foo', return_ctx=True)
    # yes '..' is a valid tag name in Mercurial, no wonder Git doesn't like it
    wrapper.command('tag', b'..', rev=b'.')

    handler = HeptapodGitHandler(wrapper.repo, wrapper.repo.ui)
    # just enough so that the changeset seems to be known on the Git side
    handler.map_set(b'should be a Git SHA', ctx.hex())
    changes = handler.heptapod_compare_tags({})
    assert changes == {}



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

from mercurial import pycompat


def switch_user(wrapper, user):
    """Common test helper to change the name of the acting user.

    This helps enforcing that it works uniformly for all permission
    related hooks.
    """
    wrapper.repo.ui.environ[b'REMOTE_USER'] = pycompat.sysbytes(user)



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Tests for hooks that don't need neither a repo, nor to fake requests.
"""
import pytest

from ..hooks import (
    format_alert_message,
    format_warnings,
    format_mr_links,
    git_hook_format_changes,
    parse_broadcast_msg,
    parse_actor,
    InvalidActor,
)


ZERO_SHA = b"0" * 20


def test_parse_actor():
    assert parse_actor("key-1") == ('key_id', '1')
    assert parse_actor("key-17") == ('key_id', '17')
    assert parse_actor("user-3") == ('user_id', '3')
    assert parse_actor("user-35") == ('user_id', '35')
    assert parse_actor("username-testgr") == ('username', 'testgr')
    with pytest.raises(InvalidActor) as exc_info:
        parse_actor('key-1x')
    assert exc_info.value.args[0] == 'key-1x'
    with pytest.raises(InvalidActor) as exc_info:
        parse_actor('user-1a')
    assert exc_info.value.args[0] == 'user-1a'
    with pytest.raises(InvalidActor) as exc_info:
        parse_actor('imaginary-12')
    assert exc_info.value.args[0] == 'imaginary-12'


def test_git_hook_format_changes():
    changes = {b'branch/newbr': (ZERO_SHA, b"beef1234dead" + b"a" * 8),
               b'topic/default/zz': (b"23cafe45" + b"0" * 12,
                                     b"1234" + b"b" * 16),
               }
    assert set(git_hook_format_changes(changes).splitlines()) == {
        b'00000000000000000000 beef1234deadaaaaaaaa branch/newbr',
        b'23cafe45000000000000 1234bbbbbbbbbbbbbbbb topic/default/zz'
    }


def test_parse_broadcast_msg():
    msg = "Short message"
    assert parse_broadcast_msg(msg, 80) == [msg]

    msg = "A somewhat longer message, cut here"
    assert parse_broadcast_msg(msg, 30) == ["A somewhat longer message,",
                                            "cut here"]


def test_format_alert_message():
    # here the point of the test is mostly to make sure that we don't
    # get stupid errors, we don't care that much of the exact amount of
    # padding or leading and trailing separators
    msg = "Short message"
    assert (format_alert_message(msg)[3]
            == b'                             Short message'
            )
    msg = ' '.join('word%d' % i for i in range(20))
    formatted = format_alert_message(msg)
    assert (formatted[3]
            == (b'   word0 word1 word2 word3 word4 word5 word6 word7 word8 '
                b'word9 word10')
            )
    assert (formatted[4]
            == (b'     word11 word12 word13 word14 word15 '
                b'word16 word17 word18 word19')
            )


def test_format_warnings():
    """Test format_warnings for non breakage, not for precise content wording.
    """
    assert b'WARNINGS' in format_warnings('Short warning')[3]


def fmt_mr_links(mrs):
    """Call and convert return into a tuple.

    The goal is that assertions don't depend onto the precise return type.
    """
    return tuple(format_mr_links(mrs))


def test_format_mr_links():
    """Test format_mr_links for non breakage, not for precise content wording.
    """
    assert fmt_mr_links(None) == ()
    assert fmt_mr_links([]) == ()
    assert fmt_mr_links(()) == ()

    seen_create = False
    seen_view = False

    for line in fmt_mr_links(
            [dict(new_merge_request=True,
                  branch_name='heptapod-next',
                  url="https://heptapod.example?query=something"),
             dict(new_merge_request=False,
                  branch_name='heptapod-stable',
                  url="https://heptapod.example?query=other"),
             ]):
        line = line.lower()
        if b'view merge request' in line:
            seen_view = True
            assert b"heptapod-stable" in line
            # query string passed through
            assert b'?query=other' in line
        elif b'create a merge request' in line:
            seen_create = True
            assert b"heptapod-next" in line
            assert b'?query=something' in line
    assert seen_view, "Did not get message to view existing MR"
    assert seen_create, "Did not get message to create new MR"



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import

import pytest
from heptapod.testhelpers import LocalRepoWrapper
from mercurial import pycompat

from .utils import common_config

parametrize = pytest.mark.parametrize


def make_repo(tmpdir, config):
    full_config = common_config()
    full_config.update(config)
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'), config=full_config)
    return wrapper.repo.ui


def test_multiple_heads_allow(tmpdir):
    ui = make_repo(tmpdir,
                   config=dict(heptapod={'allow-multiple-heads': 'yes'},
                               experimental={'single-head-per-branch': 'no'},
                               ))
    assert not ui.configbool(b'experimental', b'single-head-per-branch')


def test_bookmarks_allow(tmpdir):
    ui = make_repo(tmpdir,
                   config=dict(heptapod={'allow-bookmarks': 'yes'},
                               ))
    assert not ui.configbool(b'experimental', b'single-head-per-branch')
    assert ui.configbool(b'experimental',
                         b'hg-git.bookmarks-on-named-branches')


def test_auto_publish_nothing(tmpdir):
    ui = make_repo(tmpdir,
                   config=dict(
                       heptapod={'auto-publish': 'nothing'},
                       experimental={'topic.publish-bare-branch': 'yes'},
                   ))
    assert not ui.configbool(b'experimental', b'topic.publish-bare-branch')


def test_auto_publish_all(tmpdir):
    ui = make_repo(tmpdir,
                   config=dict(
                       heptapod={'auto-publish': 'all'},
                       phases={'publish': 'no'},
                   ))
    assert ui.configbool(b'phases', b'publish')


@parametrize('key',
             [('experimental', 'single-head-per-branch'),
              ('experimental', 'hg-git.bookmarks-on-named-branches'),
              ('experimental', 'topic.publish-bare-branch'),
              ('phases', 'publish'),
              ])
@parametrize('value', [True, False])
def test_lower_level_config_untouched(tmpdir, key, value):
    """This case is the default config in Heptapod normal conditions."""
    section, item = key
    ui = make_repo(tmpdir, config={section: {item: value}})
    assert ui.configbool(pycompat.sysbytes(section),
                         pycompat.sysbytes(item)) == value



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

from hgext3rd.heptapod import hggit_parse_hgsub
from hggit.util import OrderedDict


def parse_hgsub(lines):
    return hggit_parse_hgsub(None, lines)


def test_parse_hgsub():
    assert parse_hgsub([b""]) == OrderedDict()
    parsed = parse_hgsub((b"foo2=bar2\n", b"foo=bar"))
    assert isinstance(parsed, OrderedDict)
    assert parsed == OrderedDict(((b"foo2", b"bar2"),
                                  (b"foo", b"bar")))

    # [subpaths] section is simply ignored, in particular we don't
    # barf on it (see heptapod#310)
    assert parse_hgsub((b"foo=bar\n",
                        b"[subpaths]\n",
                        b"foo2=bar\n")) == OrderedDict([(b"foo", b"bar")])



class HeadPruneReason(object):
    """Common base class for reasons a GitLab branch would disappear.

    A GitLab branch corresponds either a branch/topic combination, or to
    a named branch, or to a bookmark.

    The bundled information is important for Merge Request detection on the
    GitLab side.

    :attr:`key` is to be used for extraction.
    :attr:`info` complementary information that can be crucial for
          Merge Request detection or any other action on the GitLab side.

    This base class can be used for an unspecified reason::

        >>> reason = HeadPruneReason()
        >>> reason
        HeadPruneReason(None)

    Extraction without further details for a ref just gives the ref back
    (actually on Python3, ref will be bytes)::

        >>> reason.extract() is None
        True
    """
    key = None

    def __init__(self, info=None):
        self.info = info

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.info)

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.info == other.info

    def extract(self):
        """Extract information for serialization in GitLab API message.

        Can be overridden in subclasses
        """
        return self.info


class HeadPruneReasonWithSha(HeadPruneReason):
    """Base class for all prune reasons whose info is a single SHA (bytes)
    """

    def extract(self):
        return self.info.decode()


class TopicPublished(HeadPruneReasonWithSha):
    """The current transaction published the topic GitLab branch.

    :attr:`info` is the GitLab SHA of the last public changeset
          bearing the given branch/topic combintation. It is useful for
          Merge Request updates, so that they display the same thing as
          if the changeset and their publication had been pushed separately.

    ::

        >>> reason = TopicPublished(b'last_sha')
        >>> 'last_sha' in repr(reason)
        True
        >>> print(reason.extract())
        last_sha
    """
    key = 'topic_published'


class BranchClosed(HeadPruneReason):
    """The GitLab branch is closed within the current transaction.

    :attr:`info` is a list of pairs, one for each new closing changeset.
         The first element of each pair is the GitLab SHA of the new closing
         changeset, and the second element is the list of GitLab SHAs for its
         parents. Both information can be useful in Merge Request detection.
         Can also be ``None`` in cases of lingering old closed branches.

    >>> BranchClosed().extract() is None
    True
    """
    key = 'branch_closed'

    def extract(self):
        if self.info is None:
            return None
        return [[closing.decode(), [p.decode() for p in parents]]
                for closing, parents in self.info]


class BookmarkRemoved(HeadPruneReason):
    """
    :attr:`info` is the GitLab SHA for the previous position of the bookmark
    """
    key = 'bookmark_removed'


class WildHeadResolved(HeadPruneReasonWithSha):
    """Used when a wild head is no more a head.

    Could be because it has more wild descendents, or because it's been
    properly merged.
    """
    key = 'wild_resolved'


class AllHeadsBookmarked(HeadPruneReason):
    """Used if all heads of a named branch become bookmarked.

    In that case, the GitLab branch for the named branch simply disappears.
    """
    key = 'all_heads_bookmarked'



# Copyright 2019 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Helpers for automatic tests.

These allow both high level operation on testing repos, and lower level
calls and introspections, making it possible to test more exhaustively inner
code paths that with `.t` tests, which are really functional tests.
"""
import os
from mercurial import (
    cmdutil,
    commands,
    hg,
    node,
    phases,
    ui as uimod,
)
import random
import time

# re-exports for stability
NULL_REVISION = node.nullrev
NULL_ID = node.nullid


def as_bytes(s):
    """Whether s is a path, an str, or bytes, return a bytes representation.

    Explicitely, the conversion to bytes is done in UTF-8, so that it cannot
    fail.

    If that doesn't suit your case, just encode your string or path
    before hand as needed.
    """
    if s is None:
        return None
    if isinstance(s, bytes):
        return s
    return str(s).encode('utf-8')


def make_ui(base_ui, config=None):
    # let's make sure we aren't polluted by surrounding settings
    os.environ['HGRCPATH'] = ''
    if base_ui is None:
        ui = uimod.ui.load()
    else:
        ui = base_ui.copy()

    # with load(), ui.environ is encoding.environ, with copy() it's not copied
    # we need the environ to be unique for each test to avoid side effects.
    # Also, on Python 2, encoding.environ is os.environ, leading to even
    # worse side effects.
    ui.environ = dict(ui.environ)

    # we want the most obscure calls to ui output methods to be executed
    # in tests. With the default settings, ui.note() early returns for example
    # meaning that only its arguments are tested, not what it does of them.
    ui.setconfig(b'ui', b'debug', b'yes')

    if config is not None:
        for section_name, section in config.items():
            for item_name, item_value in section.items():
                ui.setconfig(as_bytes(section_name),
                             as_bytes(item_name),
                             as_bytes(item_value),
                             source=b'tests')
    return ui


class LocalRepoWrapper(object):
    """Facilities for handling Mercurial repositories.

    As the name suggests, this is a wrapper class that embeds an
    instance of :class:`mercurial.localrepo.localrepo` as :attr:`repo`.

    It provides helper methods for initialization and content creation or
    mutation, both in the working directory and in changesets.

    For convenience, these high level methods accept both unicode and bytes
    strings. The path objects used by pytest can be used readily where
    a path is expected. There is one notable exception to this:
    the generic :meth:`command` which is designed to forward its arguments
    to the underlying Mercurial command directly.

    All return values that are from Mercurial are untouched, hence strings
    would be bytes.

    All conversions to bytes are done with the UTF-8 codec. If that doesn't
    suit your case, simply encode your strings before hand.
    Running unmodified tests based on this helpers on a non UTF-8 filesystem
    is not supported at this point, but it could be done, we are open to
    suggestions.
    """

    def __init__(self, repo):
        self.repo = repo

    @classmethod
    def init(cls, path, base_ui=None, config=None):
        path = as_bytes(str(path))
        init = cmdutil.findcmd(b'init', commands.table)[1][0]
        ui = make_ui(base_ui, config)
        init(ui, dest=path)
        return cls(hg.repository(ui, path))

    @classmethod
    def load(cls, path, base_ui=None, config=None):
        ui = make_ui(base_ui, config=config)
        return cls(hg.repository(ui, as_bytes(path)))

    @classmethod
    def share_from_path(cls, src_path, dest_path,
                        ui=None, base_ui=None, config=None,
                        **share_opts):
        """Create a new repo as the `share` command would do.

        :param ui: if specified, will be copied and used for the new repo
                   creation through ``share``
        :param config: only if ``ui`` is not specified, will be used to
                       create a new ``ui`` instance
        :param base_ui: only if ``ui`` is not specified, will be used to
                       create a new :class:`ui` instance
        :param share_opts: passed directly to :func:`hg.share()`
        :return: wrapper for the new repo
        """
        if ui is None:
            ui = make_ui(base_ui, config=config)
        else:
            # TODO not enough for environ independence
            ui = ui.copy()

        # the 'share' command defined by the 'share' extension, is just a thin
        # wrapper around `hg.share()`, which furthermore returns a repo object.
        return cls(hg.share(ui, as_bytes(src_path), dest=as_bytes(dest_path),
                            **share_opts))

    def share(self, dest_path, **share_opts):
        return self.share_from_path(self.repo.root, dest_path,
                                    ui=self.repo.ui, **share_opts)

    def command(self, name, *args, **kwargs):
        cmd = cmdutil.findcmd(as_bytes(name), commands.table)[1][0]
        repo = self.repo
        return cmd(repo.ui, repo, *args, **kwargs)

    def write_commit(self, rpath,
                     content=None, message=None,
                     return_ctx=True,
                     parent=None,
                     branch=None,
                     user=None,
                     utc_timestamp=None,
                     topic=None):
        """Write content at rpath and commit in one call.

        This is meant to allow fast and efficient preparation of
        testing repositories. To do so, it goes a bit lower level
        than the actual commit command, so is not suitable to test specific
        commit options, especially if through extensions.

        This leaves the working directoy updated at the new commit.

        :param rpath: relative path from repository root. If existing,
                      will be overwritten by `content`
        :param content: what's to be written in ``rpath``.
                        If not specified, will be replaced by random content.
        :param message: message commit. If not specified, defaults to
                        ``content``
        :param parent: binary node id or :class:`changectx` instance.
                       If specified, the repository is
                       updated to it first. Useful to produce branching
                       histories. This is single valued, because the purpose
                       of this method is not to produce merge commits.
        :param user: full user name and email, as in ``ui.username`` config
                     option. Can be :class:`str` or :class:`bytes`
        :param utc_timestamp: seconds since Epoch UTC. Good enough for
                              tests without ambiguity. Can be float (only
                              seconds will be kept). Defaults to
                              ``time.time()``
        :param return_ctx: if ``True``, returns a :class:`changectx` instance
                           and a binary node id otherwise, which can be more
                           straightforward and faster in some cases.
        :returns: :class:`changectx` instance or binary node id for the
                  generated commit, according to ``return_ctx``.
        """
        repo = self.repo
        path = os.path.join(repo.root, as_bytes(rpath))
        if parent is not None:
            if isinstance(parent, bytes):
                self.update_bin(parent)
            else:
                self.update(parent.hex())

        if content is None:
            content = "random: {}\n\nparent: {}\n".format(
                random.random(),
                node.hex(repo.dirstate.p1()))
        content = as_bytes(content)

        if message is None:
            message = content

        if branch is not None:
            self.repo.dirstate.setbranch(as_bytes(branch))

        if topic is not None:
            self.command(b'topics', as_bytes(topic))

        with open(path, 'wb') as fobj:
            fobj.write(content)

        if utc_timestamp is None:
            utc_timestamp = time.time()

        def commitfun(ui, repo, message, match, opts):
            return self.repo.commit(message,
                                    as_bytes(user),
                                    (int(utc_timestamp), 0),
                                    match=match,
                                    editor=False,
                                    extra=None)
        new_node = cmdutil.commit(repo.ui, repo, commitfun, (path, ),
                                  {b'addremove': True,
                                   b'message': as_bytes(message),
                                   })
        return repo[new_node] if return_ctx else new_node

    def update_bin(self, bin_node, **opts):
        """Update to a revision specified by its node in binary form.

        This is separated in order to avoid ambiguities
        """
        # maybe we'll do something lower level later
        self.update(node.hex(bin_node), **opts)

    def update(self, rev, hidden=False):
        repo = self.repo.unfiltered() if hidden else self.repo
        cmdutil.findcmd(b'update', commands.table)[1][0](repo.ui, repo,
                                                         as_bytes(rev))

    def set_phase(self, phase_name, revs, force=True):
        repo = self.repo
        opts = dict(force=force, rev=[as_bytes(r) for r in revs])
        phase_name_bytes = as_bytes(phase_name)
        opts.update((phn.decode(), phn == phase_name_bytes)
                    for phn in phases.cmdphasenames)
        cmdutil.findcmd(b'phase', commands.table)[1][0](repo.ui, repo, **opts)

    def prune(self, revs, successors=(), bookmarks=()):
        # the prune command expects to get all these arguments (it relies
        # on the CLI defaults but doesn't have any at the function call level).
        # They are unconditionally concatened to lists, hence must be lists.
        # (as of Mercurial 5.3.1)
        if isinstance(revs, (bytes, str)):
            revs = [revs]
        return self.command('prune',
                            rev=[as_bytes(r) for r in revs],
                            new=[],  # deprecated yet expected
                            # TODO py3 convert these two to bytes as needed:
                            successor=list(successors),
                            bookmark=list(bookmarks))



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import
from heptapod.testhelpers import (
    LocalRepoWrapper,
    make_ui,
)
from mercurial import pycompat
import os

from .utils import common_config


def test_hook(tmpdir):
    """Heptapod environ variables are passed over in hooks."""
    out_path = tmpdir.join('out')
    config = common_config()
    config['hooks'] = dict(commit="echo $HEPTAPOD_VARIABLE > %s" % out_path)
    wrapper = LocalRepoWrapper.init(tmpdir.join('repo'), config=config)
    wrapper.repo.ui.environ[b'HEPTAPOD_VARIABLE'] = b'hepta-value'
    wrapper.write_commit("foo")
    assert out_path.read() == 'hepta-value\n'


def test_none():
    """Test exceptional call with no environment."""
    out = pycompat.bytesio()
    ui = make_ui(None)
    ui.environ[b'HEPTAPOD_VARIABLE'] = b'hepta-val'
    ui._runsystem(cmd=b"echo -n $HEPTAPOD_VARIABLE",
                  environ=None,
                  cwd=os.getcwd(),
                  out=out)
    assert out.getvalue() == b'hepta-val'
